'''These are variables that are useful to reference.
There are several curated lists of link shortening domains,
the url of datasets for tutorials.
'''

__author__= 'Leon Yin'

headers = {'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/35.0.1916.47 Safari/537.36'}

# these domains need a redirect (Leon Yin + UnshortenIT 2018)
short_domain_ad_redirects = [
    'sh.st',  'adf.ly', 'lnx.lu',
    'adfoc.us', 'j.gs','q.gs',
    'u.bb', 'ay.gy', 'atominik.com',
    'tinyium.com', 'microify.com', 'linkbucks.com',
    'www.linkbucks.com', 'jzrputtbut.net', 'any.gs',
    'cash4links.co', 'cache4files.co', 'dyo.gs',
    'filesonthe.net', 'goneviral.com', 'megaline.co',
    'miniurls.co', 'qqc.co', 'seriousdeals.net',
    'theseblogs.com', 'theseforums.com',
    'tinylinks.co', 'tubeviral.com', 'ultrafiles.net',
    'urlbeat.net', 'whackyvidz.com','yyv.co',
    'href.li', 'anonymz.com', 'festyy.com', 
    'ceesty.com', 'tiny.cc'
]

# these are standard domain shorteners (Leon Yin 2018)
short_domain = [
    'dlvr.it', 'bit.ly', 'buff.ly',
    'ow.ly',  'goo.gl', 'shar.es', 
    'ift.tt', 'fb.me',  'washex.am', 
    'smq.tc',  'trib.al', 'is.gd',
    'paper.li', 'waa.ai', 'tinyurl.com',
    'ht.ly', '1.usa.gov', 'deck.ly',
    'bit.do', 'lc.chat', 'urls.tn',
    'soo.gd',  's2r.co', 'clicky.me',
    'budurl.com', 'bc.vc', 'branch.io',
    'capsulink.com', 'ux9.de', 'fuck.it',
    't2m.io', 'shrt.li', 'elbo.in',
    'shrtfly.com', 'hiveam.com',
    'slink.be', 'plu.sh', 'cutt.ly',
    'zii.bz', 'munj.pw', 
]

# there are domain shorteners for common news outlets (Leon Yin 2018).
short_domain_media = [
    'on.rt.com', 'wapo.st', 'hill.cm', 
    'dailym.ai', 'cnn.it', 'nyti.ms',
    'politi.co', 'fxn.ws', 'usat.ly', 
    'huff.to', 'nyp.st', 'cbsloc.al',
    'wpo.st', 'on.wsj.com', 'nydn.us',
]

all_short_domains = short_domain_ad_redirects + short_domain + ['ln.is']

congress_dataset_url = ('https://raw.githubusercontent.com/SMAPPNYU/'
                        'urlExpander/master/datasets/'
                        'congress_sample_links.csv')

us_nation_domain_url = ('https://raw.githubusercontent.com/SMAPPNYU/'                             'urlExpander/master/datasets/'
                        'us_national_domains.csv')


"""This module has utility functions for parsing links from Tweets.
Check out the smappdragon package for Tweet parsing.
https://github.com/SMAPPNYU/smappdragon
"""

from urlexpander.core.api import get_domain

__all__ = ['get_link']
__author__= 'Leon Yin'

def _get_full_text(tweet):
    '''
    Parses a tweet json to retrieve the full text.
    
    :input tweet: a nested dictionary of a Tweet either from the streaming or search API.
    :returns: string of the full_text field hidden in the Tweet
    '''
    if "extended_tweet" in tweet and "full_text" in tweet["extended_tweet"]:
        return tweet["extended_tweet"]["full_text"]
    try:
        return tweet["text"]
    except:
        return tweet['full_text']
    
def get_link(tweet):
    '''
    Returns a generator containing tweet metadata about media.
    
    The metadata dict contains the following columns:
    
    columns = {
      'link.domain' : 'the domain of the URL', 
      'link.url_long' : 'the URL (this can be short!)', 
      'link.url_short' : 'The t.co URL', 
      'tweet.created_at' : 'When the tweet was created', 
      'tweet.id' : 'The ID of the tweet', 
      'tweet.text' : 'The Full text of the tweet', 
      'user.id' : 'The Twitter ID of the tweeter'
    }
    
    :input tweet: a nested dictionary of a Tweet either from the streaming or search API.
    :returns: a generator of dictionaries
    '''
    if not isinstance(tweet, dict):
        return
    
    row = {
        'user_id': tweet['user']['id'],
        'tweet_id': tweet['id'],
        'tweet_created_at': tweet['created_at'],
        'tweet_text' : _get_full_text(tweet)
    }

    list_urls = tweet['entities']['urls']
    
    if list_urls:
        for url in list_urls:
            r = row.copy()
            r['link_url_long'] = url.get('expanded_url')
            
            if r['link_url_long']:
                r['link_domain'] = get_domain(r['link.url_long'])
                r['link_url_short'] = url.get('url')

                yield r  

                
def count_matrix(df, user_col='user.id', domain_col='link.domain', 
                 unique_count_col='tweet.id', domain_list=[]):
    '''
    Creates a count matrix of number of domains shared per user. 
    Where each column is a count of domains, and each row represents on user.
    
    
    :input df: (pandas dataframe) an un-aggregrated dataframe of links shared by user.
    :input user_col: (str) the name of the column in input dataframe to aggragate on. 
            Feeds into the index argument in `pd.pivot_table`.
    :inout domain_col: (str) the name of the column in the input dataframe to count.
            Feeds into the columns argument in `pd.pivot_table`.
    :input count_unique_col: (str) the name of the column to count unique values amongst domain_col.
    :input domain_list: (list) a list of normalized domains to create the count matrix on.
    
    :returns: matrix (pandas dataframe) counts per domain by user.
    '''
    # check the correct columns are present
    for col in [user_col, domain_col, unique_count_col]:
        try:
            assert(col in df.columns)
        except:
            raise ValueError(f"{col} is not a column in the input dataframe")
    
    # what are all the domains available?
    all_domains = df[domain_col].unique()    
    
    # create a count matrix
    matrix = df.pivot_table(index=user_col, 
                            columns=[domain_col],
                            values=[unique_count_col],
                            aggfunc=lambda x: len(x.unique()),
                            fill_value=0)
    
    # standardize the column names, and remove any hierarchys for readability.
    matrix = matrix.T.reset_index(level=0, drop=True).T
    matrix.columns.name = None
    
    # filter out columns not included in domain_list
    if domain_list:
        matrix = matrix[[c for c in all_domains if c in domain_list]]
    
    return matrix




import pandas as pd

from urlexpander.core import constants

__all__ = ['congress_twitter_links']
__author__= 'Leon Yin'

def load_congress_twitter_links():
    '''
    Returns a Pandas dataframe of a random sample of 50K 
    URLs parsed from Twitter aaccounts from the 115th congress.
    
    URLs are parsed using `tweet_parser.get_link`.
    
    columns = {
      'link.domain' : 'the domain of the URL', 
      'link.url_long' : 'the URL (this can be short!)', 
      'link.url_short' : 'The t.co URL', 
      'tweet.created_at' : 'When the tweet was created', 
      'tweet.id' : 'The ID of the tweet', 
      'tweet.text' : 'The Full text of the tweet', 
      'user.id' : 'The Twitter ID of the tweeter'
    }
    
    :returns: (df) of Tweets with cols 
    '''
    congress = pd.read_csv(constants.congress_dataset_url,
                              dtype={'tweet.id':str,'user.id':str})
    return congress


def load_us_national_domains():
    '''
    Returns a Pandas Dataframe of top-level domains
    of national media sites.
    
    :returns: (df) of domains
    '''
    us_national_domains = pd.read_csv(constants.us_nation_domain_url)
    
    return us_national_domains


"""
This is the main module of the urlExpander package.
It houses functions to standardize urls, and extract domain names from links.
It also has the expand and multithread expand functions, which are the crux of this package.
"""

import os
import glob
import json
import itertools
import datetime
import concurrent
import requests
from urllib.parse import urlparse

import unshortenit
from tqdm import tqdm as tqdm
import numpy as np
import pandas as pd
import tldextract

from urlexpander.core import constants

__all__ = ['strip_url',
           'get_domain',
           'is_short',
           'is_short_domain',
           'expand',
           'multithread_expand',
           'count_matrix']

__author__ = 'Leon Yin'

def strip_url(url):
    '''
    Best attempt to standardize websites.
    
    :input url: a string of a URl
    :returns: a URL stripped of www. and https:
    '''
    if url:
        return url.replace('www.', '').replace('http://', '').replace('https://', '').rstrip('/')
    else:
        return url

def _chunks(l, chunksize):
    """Yield successive n-sized chunks from l.
    Taken from https://stackoverflow.com/a/312464/5094480
    """
    for i in range(0, len(l), chunksize):
        yield l[i:i + chunksize]

def get_domain(url):
    '''
    Returns domain name of a URL (and removes "www.")
    e.g. 'https://www.nytimes.com/2016/12/23/upshot/...' to 'nytimes.com'
    
    :input url: a string of a url, can contain https://
    :returns: a string of the domain
    '''
    extracted = tldextract.extract(url)
    if extracted.suffix == '' or extracted.domain == '':
        domain = url
    elif extracted.subdomain == '':
        domain = "{}.{}".format(extracted.domain, extracted.suffix)
    else:
        domain = "{}.{}.{}".format(extracted.subdomain, extracted.domain, extracted.suffix)
    domain = domain.lower().replace('www.', '')
    
    return domain

def is_short(url, list_of_domains=constants.all_short_domains):
    '''
    A function which returns True if a domain is in a list_of_domains. 
    Make sure that domain and list_of_domains is preprocessed (or not at all),
    in the same way.
    
    :input domain: (str) a standardized domain name
    :input list_of_domains: (list of str) of standardized domain names
    :returns:
    '''
    domain = get_domain(url)
    

    return is_short_domain(domain, list_of_domains=list_of_domains)

def is_short_domain(domain, list_of_domains=constants.all_short_domains):
    '''
    A function which returns True if a domain is in a list_of_domains. 
    Make sure that domain and list_of_domains is preprocessed (or not at all),
    in the same way.
    
    :input domain: (str) a standardized domain name
    :input list_of_domains: (list of str) of standardized domain names
    :returns:
    '''
    if domain in list_of_domains:
        return True
    return False

def _parse_error(error):
    '''
    Although some redirects no longer work, we can still use the response from the error to figure out where it would have gone.
    This function parses error messages from requests.head (called in expand), to try to figure out what website the bit-link was intended to re-direct to.
    
    :param error: a string of the error response from unshorten.
    :returns: the domain parsed from the error string, and the long_url. If the error can't be parsed, the domain is -1
    '''
    if 'ConnectionPool' in error:
        vals = error.split("ConnectionPool(host='")[1].split("',")
        domain = vals[0]
        url_endpoint = vals[1].split('Max retries exceeded with url: ')[-1].split(" (Caused by")[0]
        url_endpoint = os.path.join('http://', domain + url_endpoint)
        
    elif 'Client Error: ' in error or 'Server Error' in error:
        url_endpoint = error.split(" for url: ")[-1]
        domain = get_domain(url_endpoint)
        
    else:
        domain, url_endpoint = -1, error
    
    return domain, url_endpoint


def expand(link, timeout=2, **kwargs):
    '''
    Expands a url, while taking into consideration: special link shortener or analytics platforms that either need a sophisticated
    redirect(st.sh), or parsing of the url (ln.is)
    
    :param link: string of a link to unshorten.
    :returns: a dictionary with the original link, the unshortened link, and the unshortened domain.
    '''
    try:
        r = requests.head(link, 
                            allow_redirects=True, 
                            timeout=timeout,
                            **kwargs)
        r.raise_for_status()
        url_long = r.url
        domain = get_domain(url_long)
        
    except requests.exceptions.RequestException as e:
        domain, url_long = _parse_error(str(e))

    if domain == 'ln.is':
        url_long = link.replace('ln.is/', '')
        domain = get_domain(url_long)
        
    elif domain in constants.short_domain_ad_redirects or domain == -1:
        url_long = unshortenit.UnshortenIt().unshorten(link,
                                                       timeout=timeout)
        domain = get_domain(url_long)

    return dict(original_url=link,
                resolved_domain=domain,
                resolved_url=url_long)


def multithread_expand(links_to_unshorten, chunksize=1280, n_workers=64, 
                       cache_file=None, random_seed=303, 
                       return_errors=True, **kwargs):
    '''
    Calls unshorten_3 with multiple (n_workers) threads to unshorten a list of urls.
    
    :param links_to_unshorten: (list) a list of urls (str) to unshorten
    :param chunksize: (int) chunks links_to_unshorten, which makes computation quicker with larger inputs
    :param n_workers: (int) how many threads
    :param cache_file: (str) a path to a json file to read and write results in
    :param random_seed: (int) initializes the random state for shuffling the input
    :param return_errors: (bool) whether to return a tuple of dataframe of shortened link, and dict of errors
        or just the dataframe of shortened links
        
    
    :returns: (list) a list of dictionaries perfect for Pandas Dataframes.
    '''
    
    # shuffle the inputs, this is to reduce the changes of making requests to the same domain.
    np.random.seed(random_seed)
    np.random.shuffle(links_to_unshorten)

    # read cache file
    unshortened_urls = []
    error = []
    if cache_file and os.path.exists(cache_file):
        with open(cache_file, 'r') as f_:
            for line in f_:
                unshortened_urls.append(json.loads(line))
            abd_ = [_['original_url'] for _ in unshortened_urls]
            links_to_unshorten = [link for link in links_to_unshorten if link not in abd_]
    
    # chunk the list of arguments
    for chunk in tqdm(_chunks(links_to_unshorten, chunksize=chunksize)):
        # create n_workers threads, and map chunked argumnets to them
        with concurrent.futures.ThreadPoolExecutor(max_workers = n_workers) as executor:
            future_to_url = {executor.submit(expand, url, **kwargs): 
                             url for url in chunk}
            for i, future in enumerate(concurrent.futures.as_completed(future_to_url)):
                try:
                    data = future.result()
                except Exception as exc:
                    data = str(type(exc))
                    error.append({chunk[i] : str(type(exc))})
                finally:
                    unshortened_urls.append(data)
                    # save the results
                    if cache_file and isinstance(data, dict):
                        with open(cache_file, 'a') as f_:
                            f_.write(json.dumps(data) + '\n')
    
    if return_errors:
        return unshortened_urls, error
    else:
        return unshortened_urls
    


import sys
from setuptools import setup

long_description = """This package makes working with link data from social media and webpages easier. It not only expands links, but catches errors, and makes parallel link expansion quick and efficient.
```
import urlexpander as ux
ux.expand('https://trib.al/xXI5ruM')
```
returns
```
{'original_url': 'https://trib.al/xXI5ruM',
 'resolved_domain': 'breitbart.com',
 'resolved_url': 'https://www.breitbart.com/video/2017/12/31/lindsey-graham-trump-just-cant-tweet-iran/'}
 ```
 
 Please take a look at the [quickstart](http://nbviewer.jupyter.org/github/SMAPPNYU/urlExpander/blob/master/examples/quickstart.ipynb?flush=true).

    """

setup(
    name="urlexpander",
    packages=['urlexpander', 'urlexpander.core'],
    py_modules=['urlexpander'],
    version='0.0.26',
    description="urlExpander is a Python package for quickly and thoroughly expanding shortened URLs.",
    long_description=long_description,
    author="leon yin",
    author_email="whereisleon@gmail.com",
    url="https://github.com/SMAPPNYU/urlExpander",
    keywords='smapp social media unshorten expand link url',
    license="MIT",
    install_requires=[
        'tldextract',
        'pandas',
        'numpy',
        'tqdm',
        'unshortenit'
    ]
)



"""This module has utility functions for scraping text from URLs,
Specifically it helps get the title, description (like what shows up on Google), and paragraphs.
"""

import re
import html
import requests
from collections import OrderedDict

from urlexpander.core import constants

__all__ = ['get_webpage_meta',
           'get_webpage_paragraphs',
           'get_webpage_description',
           'get_webpage_title']
__author__= 'Leon Yin'


def _search_webpage_title(text, headers=constants.headers, **kwargs):
    title = None
    regex = re.compile('<title>(.*?)</title>', re.IGNORECASE|re.DOTALL)
    try:
        title = regex.search(text).group(1)
        title = html.unescape(title)
    except: pass
    return title

    
def _search_webpage_description(text, headers=constants.headers, **kwargs):
    desc = None
    regex =  re.compile('<meta property="og?:description" content="(.*?)>', re.IGNORECASE|re.DOTALL)
    try:
        desc = regex.search(text).group(1).rstrip('/').rstrip(' ').rstrip('"')
        desc = html.unescape(desc)
    except:
        pass
    return desc

def _search_webpage_paragraphs(text, headers=constants.headers, **kwargs):
    paragraphs = []
    try:
        paragraphs = re.findall(r'<p>(.*?)</p>', text)
        paragraphs = [html.unescape(p) for p in paragraphs]
    except:
        pass
    return paragraphs

def get_webpage_title(url, headers=constants.headers, **kwargs):
    r = requests.get(url, headers=headers, **kwargs)
    title = _search_webpage_title(r.text)
    return title

def get_webpage_description(url, headers=constants.headers, **kwargs):
    r = requests.get(url, headers=headers, **kwargs)
    desc = _search_webpage_description(r.text)
    return desc

def get_webpage_paragraphs(url, headers=constants.headers, **kwargs):
    r = requests.get(url, headers=headers, **kwargs)
    paragraphs = _search_webpage_paragraphs(r.text)
    return paragraphs

def get_webpage_meta(url, headers=constants.headers, **kwargs):
    r = requests.get(url, headers=headers, **kwargs)
    text = r.text
    meta = OrderedDict(url = url,
                       title = _search_webpage_title(text),
                       description = _search_webpage_description(text),
                       paragraphs = _search_webpage_paragraphs(text))
    return meta



from urlexpander.core.api import(
    strip_url,
    get_domain,
    is_short,
    is_short_domain,
    expand,
    multithread_expand,
)

from urlexpander.core import (
    tweet_utils, 
    html_utils, 
    constants, 
    datasets
)

__version__ = '0.0.26'
__author__ = 'Leon Yin'